/**
 * Copyright (c) 2010 Alois Glomann gnalff@lucubrators.net
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package net.lucubrators.honeycomb.defaultimpl.conversion;

import java.lang.reflect.Array;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Set;

import net.lucubrators.honeycomb.defaultimpl.conversion.exceptions.ConversionException;
import net.lucubrators.honeycomb.defaultimpl.util.GenericType;

/**
 * Default implementation of a converter
 * By extending you can add own functionallity
 * 
 * @author aglomann
 * 
 */
public class HoneycombConverter {

	/**
	 * Method to convert values to other types
	 * 
	 * @param <T>
	 *            Target type of conversion
	 * @param value
	 *            the source value
	 * @param toType
	 *            the type to convert to
	 * @param reflectionType
	 *            {@link GenericType} holding additionally values about
	 *            generics
	 * @param locale
	 *            the current locale
	 * @return the converted object in the desired type
	 * @throws ConversionException
	 *             if anything went wrong, do not throw any other exception if
	 *             possible.
	 */
	@SuppressWarnings({ "rawtypes", "unchecked" })
	public <T> T convert(final Object value, final Class<?> toType, final Type reflectionType, final Locale locale) throws ConversionException {
		Object tempValue = value;
		Object result = null;

		if (tempValue == null) {
			return null;
		}
		try {
			if (!tempValue.getClass().isArray() && toType.isArray()) {
				Object temp = tempValue;
				tempValue = Array.newInstance(tempValue.getClass(), 1);
				Array.set(tempValue, 0, temp);
			}

			if (tempValue.getClass().isArray() && toType.isArray()) {
				Class<?> componentType = toType.getComponentType();

				result = Array.newInstance(componentType, Array.getLength(tempValue));

				for (int i = 0, icount = Array.getLength(tempValue); i < icount; i++) {
					Array.set(result, i, convert(Array.get(tempValue, i), componentType, reflectionType, locale));
				}
				return (T) result;
			}
			if (List.class.isAssignableFrom(toType)) {
				GenericType gt = GenericType.valueOf(reflectionType);
				if (gt.isParamterizedType()) {
					return (T) asRealTypeList(tempValue, gt.toParameterizedType(), locale);
				}
				return (T) asObjectList(tempValue);
			}
			if (Set.class.isAssignableFrom(toType)) {
				GenericType gt = GenericType.valueOf(reflectionType);
				if (gt.isParamterizedType()) {
					return (T) new HashSet(asRealTypeList(tempValue, gt.toParameterizedType(), locale));
				}
				return (T) new HashSet(asObjectList(tempValue));
			}
			if (tempValue.getClass().isArray() && !toType.isArray()) {
				tempValue = Array.get(tempValue, 0);
			}
			if ((toType == Integer.class) || (toType == Integer.TYPE)) {
				result = Integer.valueOf(stringValue(tempValue, true));

			} else if ((toType == Double.class) || (toType == Double.TYPE)) {
				result = new Double(doubleValue(tempValue));

			} else if ((toType == Boolean.class) || (toType == Boolean.TYPE)) {
				result = Boolean.valueOf(stringValue(tempValue, true));

			} else if ((toType == Byte.class) || (toType == Byte.TYPE)) {
				result = Byte.valueOf(stringValue(tempValue, true));

			} else if ((toType == Character.class) || (toType == Character.TYPE)) {
				result = stringValue(tempValue, true).toCharArray()[0];

			} else if ((toType == Short.class) || (toType == Short.TYPE)) {
				result = Short.valueOf(stringValue(tempValue, true));

			} else if ((toType == Long.class) || (toType == Long.TYPE)) {
				result = new Long(stringValue(tempValue, true));

			} else if ((toType == Float.class) || (toType == Float.TYPE)) {
				result = Float.valueOf(stringValue(tempValue, true));

			} else if (toType == BigInteger.class) {
				result = bigIntValue(tempValue);

			} else if (toType == BigDecimal.class) {
				result = bigDecValue(tempValue);

			} else if (toType == String.class) {
				result = stringValue(tempValue);

			} else if (toType == java.util.Date.class) {
				long time = longValue(tempValue);
				result = new java.util.Date(time);
			} else if (Enum.class.isAssignableFrom(toType)) {
				Class en = toType;
				result = Enum.valueOf(en, tempValue.toString());
			}
			return (T) result;
		} catch (Exception e) {
			throw new ConversionException("Could not convert " + tempValue + " to type: " + toType, e);
		}
	}

	private List<Object> asObjectList(final Object value) {
		if (value.getClass().isArray()) {
			Object[] temp = (Object[]) value;
			return Arrays.asList(temp);
		}
		return Arrays.asList(value);
	}

	@SuppressWarnings("unchecked")
	private List<?> asRealTypeList(final Object value, final ParameterizedType parameterizedType, final Locale locale) throws ConversionException {
		Type actualType = parameterizedType.getActualTypeArguments()[0];
		GenericType valueOf = GenericType.valueOf(actualType);
		if (!valueOf.isClass()) {
			throw new IllegalArgumentException("Can only convert list with class types as generic argument");
		}
		Class<?> realClass = (Class<?>) actualType;
		List<Object> asObjectList = asObjectList(value);
		@SuppressWarnings("rawtypes")
		ArrayList arrayList = new ArrayList();
		for (Object object : asObjectList) {
			arrayList.add(convert(object, realClass, parameterizedType, locale));
		}
		return arrayList;
	}

	private static long longValue(final Object value) {
		if (value == null) {
			return 0L;
		}
		Class<?> c = value.getClass();
		if (c.getSuperclass() == Number.class) {
			return ((Number) value).longValue();
		}
		if (c == Boolean.class) {
			return ((Boolean) value).booleanValue() ? 1 : 0;
		}
		if (c == Character.class) {
			return ((Character) value).charValue();
		}
		return Long.parseLong(stringValue(value, true));
	}

	private static String stringValue(final Object value, final boolean trim) {
		String result;

		if (value == null) {
			result = null;
		} else {
			result = value.toString();
			if (trim) {
				result = result.trim();
			}
		}
		return result;
	}

	private static String stringValue(final Object value) {
		return stringValue(value, false);
	}

	private static double doubleValue(final Object value) throws NumberFormatException {
		if (value == null) {
			return 0.0;
		}
		Class<?> c = value.getClass();
		if (c.getSuperclass() == Number.class) {
			return ((Number) value).doubleValue();
		}
		if (c == Boolean.class) {
			return ((Boolean) value).booleanValue() ? 1 : 0;
		}
		if (c == Character.class) {
			return ((Character) value).charValue();
		}
		String s = stringValue(value, true);

		return (s.length() == 0) ? 0.0 : Double.parseDouble(s);
	}

	private static BigInteger bigIntValue(final Object value) throws NumberFormatException {
		if (value == null) {
			return BigInteger.valueOf(0L);
		}
		Class<?> c = value.getClass();
		if (c == BigInteger.class) {
			return (BigInteger) value;
		}
		if (c == BigDecimal.class) {
			return ((BigDecimal) value).toBigInteger();
		}
		if (c.getSuperclass() == Number.class) {
			return BigInteger.valueOf(((Number) value).longValue());
		}
		if (c == Boolean.class) {
			return BigInteger.valueOf(((Boolean) value).booleanValue() ? 1 : 0);
		}
		if (c == Character.class) {
			return BigInteger.valueOf(((Character) value).charValue());
		}
		return new BigInteger(stringValue(value, true));
	}

	private static BigDecimal bigDecValue(final Object value) throws NumberFormatException {
		if (value == null) {
			return BigDecimal.valueOf(0L);
		}
		Class<?> c = value.getClass();
		if (c == BigDecimal.class) {
			return (BigDecimal) value;
		}
		if (c == BigInteger.class) {
			return new BigDecimal((BigInteger) value);
		}
		if (c.getSuperclass() == Number.class) {
			return new BigDecimal(value.toString());
		}
		if (c == Boolean.class) {
			return BigDecimal.valueOf(((Boolean) value).booleanValue() ? 1 : 0);
		}
		if (c == Character.class) {
			return BigDecimal.valueOf(((Character) value).charValue());
		}
		return new BigDecimal(stringValue(value, true));
	}
}
