/**
 * Copyright (c) 2010 Alois Glomann gnalff@lucubrators.net
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package net.lucubrators.honeycomb.defaultimpl.web.filter;

import java.io.File;
import java.io.IOException;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import net.lucubrators.honeycomb.core.bind.BinderProvider;
import net.lucubrators.honeycomb.core.engine.BindablesFactory;
import net.lucubrators.honeycomb.core.engine.HoneycombEngine;
import net.lucubrators.honeycomb.core.engine.SubmittingCombResolver;
import net.lucubrators.honeycomb.core.reflect.exceptions.CombInvocationException;
import net.lucubrators.honeycomb.defaultimpl.bind.DefaultBinderProvider;
import net.lucubrators.honeycomb.defaultimpl.conversion.HoneycombConverter;
import net.lucubrators.honeycomb.defaultimpl.engine.DefaultBindablesFactory;
import net.lucubrators.honeycomb.defaultimpl.engine.DefaultBindingMapper;
import net.lucubrators.honeycomb.defaultimpl.engine.DefaultObjectInvoker;
import net.lucubrators.honeycomb.defaultimpl.engine.DefaultSubmittingCombResolver;
import net.lucubrators.honeycomb.defaultimpl.engine.ObjectInvoker;
import net.lucubrators.honeycomb.defaultimpl.reflect.DefaultArgumentResolver;
import net.lucubrators.honeycomb.defaultimpl.reflect.DefaultMethodInvoker;
import net.lucubrators.honeycomb.defaultimpl.reflect.DefaultMethodResolver;
import net.lucubrators.honeycomb.defaultimpl.web.translator.RequestToCombIdTranslator;
import net.lucubrators.honeycomb.defaultimpl.web.translator.UriToCombIdTranslator;
import net.lucubrators.honeycomb.xml.HoneycombXMLParser;
import net.lucubrators.honeycomb.xml.constraint.ConstraintProvider;
import net.lucubrators.honeycomb.xml.constraint.DefaultConstraintProvider;
import net.lucubrators.honeycomb.xml.controller.ControllerProvider;
import net.lucubrators.honeycomb.xml.controller.DefaultControllerProvider;
import net.lucubrators.honeycomb.xml.registry.CombRegistry;
import net.lucubrators.honeycomb.xml.registry.DefaultCombRegistry;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A convinience filter for an easy setup of honeycomb
 * 
 * @author gnalFF
 * 
 */
public abstract class HoneycombFilter implements Filter {

	private static Logger log = LoggerFactory.getLogger(HoneycombFilter.class);

	private boolean chainOnNoComb;

	private HoneycombEngine honeycombEngine;

	private SubmittingCombResolver submittingCombResolver;

	private ObjectInvoker objectInvoker;

	private BinderProvider binderProvider;
	private BindablesFactory bindablesFactory;
	private HoneycombConverter honeycombConverter;

	private CombRegistry combRegistry;
	private ControllerProvider controllerProvider;
	private ConstraintProvider constraintProvider;

	private RequestToCombIdTranslator requestToCombIdTranslator;

	private FilterConfig config;

	@Override
	public void destroy() {
	}

	@Override
	public void doFilter(final ServletRequest req, final ServletResponse res, final FilterChain chain) throws IOException, ServletException {
		HttpServletRequest request = (HttpServletRequest) req;
		HttpServletResponse response = (HttpServletResponse) res;
		String translate = requestToCombIdTranslator.translate(request);

		log.trace("Translated request to combid >{}<", translate);

		boolean hasComb = combRegistry.contains(translate);
		if (!hasComb && chainOnNoComb) {
			log.trace("Found no comb, going to chain on.");
			chain.doFilter(request, response);
			return;
		}
		if (!hasComb) {
			log.trace("Found no comb. Will send a 404");
			response.setStatus(HttpServletResponse.SC_NOT_FOUND);
			return;
		}

		try {
			honeycombEngine.start(request, response, combRegistry.getById(translate));
		} catch (CombInvocationException e) {
			throw new ServletException(e.getCause());
		}
	}

	/**
	 * triggers initial setup for the {@link HoneycombEngine} and needed
	 * interfaces
	 */
	@Override
	public void init(final FilterConfig config) throws ServletException {
		this.config = config;
		customSetup();
		doDefaultSetup();
		addControllers();
		addBinders();
		addConstraints();
		scanXMLandFillRegistry(getDefinitionFile());
	}

	/**
	 * Override this method to inject custom implementation for engine
	 * interfaces
	 */
	public abstract void customSetup();

	/**
	 * Override this method to expose the file where the honeycomb definitions
	 * are
	 * 
	 * @return a file pointing to the honeycomb definitions
	 */
	public abstract File getDefinitionFile();

	/**
	 * use this method to add additional controller with
	 * {@link HoneycombFilter#addController(String, Object)}
	 * */
	public abstract void addControllers();

	/**
	 * use this method to add additional binders with the following method
	 * {@link HoneycombFilter#addBinder(String, Object)}
	 */
	public abstract void addBinders();

	/**
	 * use this method to add additional constraints with
	 * {@link HoneycombFilter#addConstraints()}
	 */
	public abstract void addConstraints();

	/**
	 * triggers default setup. Called by
	 * {@link HoneycombFilter#init(FilterConfig)}
	 */
	public void doDefaultSetup() {
		if (honeycombConverter == null) {
			honeycombConverter = new HoneycombConverter();
		}

		if (bindablesFactory == null) {
			bindablesFactory = new DefaultBindablesFactory();
		}

		if (combRegistry == null) {
			combRegistry = new DefaultCombRegistry();
		}

		if (requestToCombIdTranslator == null) {
			requestToCombIdTranslator = new UriToCombIdTranslator(".html");
		}

		if (controllerProvider == null) {
			controllerProvider = new DefaultControllerProvider();
		}
		if (constraintProvider == null) {
			constraintProvider = new DefaultConstraintProvider();
		}
		if (binderProvider == null) {
			binderProvider = new DefaultBinderProvider();
		}

		if (objectInvoker == null) {
			DefaultObjectInvoker defaultObjectInvoker = new DefaultObjectInvoker();
			defaultObjectInvoker.setArgumentResolver(new DefaultArgumentResolver());
			DefaultBindingMapper defaultBindingMapper = new DefaultBindingMapper();
			defaultBindingMapper.setHoneycombConverter(honeycombConverter);
			defaultObjectInvoker.setBindingMapper(defaultBindingMapper);
			defaultObjectInvoker.setMethodInvoker(new DefaultMethodInvoker());
			defaultObjectInvoker.setMethodResolver(new DefaultMethodResolver());
			objectInvoker = defaultObjectInvoker;
		}
		if (submittingCombResolver == null) {
			submittingCombResolver = new DefaultSubmittingCombResolver();
		}

		honeycombEngine = new HoneycombEngine();
		honeycombEngine.setBinderMethodInvoker(objectInvoker);
		honeycombEngine.setConstraintInvoker(objectInvoker);
		honeycombEngine.setControllerInvoker(objectInvoker);
		honeycombEngine.setBindablesFactory(bindablesFactory);
		honeycombEngine.setBinderProvider(binderProvider);
		honeycombEngine.setSubmittingCombResolver(submittingCombResolver);
	}

	/**
	 * scans an xml to fill up the provided {@link CombRegistry}
	 * 
	 * @param toScan
	 *            the file to scan.
	 */
	public void scanXMLandFillRegistry(final File toScan) {
		HoneycombXMLParser honeycombXMLParser = new HoneycombXMLParser(toScan);
		honeycombXMLParser.setConstraintProvider(constraintProvider);
		honeycombXMLParser.setControllerProvider(controllerProvider);
		honeycombXMLParser.setRegistry(combRegistry);
		honeycombXMLParser.scan();
	}

	public FilterConfig getFilterConfig() {
		return this.config;
	}

	public void setHoneycombConverter(final HoneycombConverter honeycombConverter) {
		this.honeycombConverter = honeycombConverter;
	}

	public void setRequestToCombIdTranslator(final RequestToCombIdTranslator requestToCombIdTranslator) {
		this.requestToCombIdTranslator = requestToCombIdTranslator;
	}

	public void setCombRegistry(final CombRegistry combRegistry) {
		this.combRegistry = combRegistry;
	}

	public void setBinderProvider(final BinderProvider binderProvider) {
		this.binderProvider = binderProvider;
	}

	public void setBindablesFactory(BindablesFactory bindablesFactory) {
		this.bindablesFactory = bindablesFactory;
	}

	public void setConstraintProvider(final ConstraintProvider constraintProvider) {
		this.constraintProvider = constraintProvider;
	}

	public void setControllerProvider(final ControllerProvider controllerProvider) {
		this.controllerProvider = controllerProvider;
	}

	public void setObjectInvoker(ObjectInvoker objectInvoker) {
		this.objectInvoker = objectInvoker;
	}

	public void setSubmittingCombResolver(SubmittingCombResolver submittingCombResolver) {
		this.submittingCombResolver = submittingCombResolver;
	}

	/**
	 * Specify if the filter should fail immediately or continue chaining, if no
	 * comb was found
	 * 
	 * @param chainOnNoComb
	 *            false if the filter should fail immediately, true if it should
	 *            chain on
	 */
	public void setChainOnNoComb(boolean chainOnNoComb) {
		this.chainOnNoComb = chainOnNoComb;
	}

	protected boolean isChainOnNoComb() {
		return chainOnNoComb;
	}

	protected SubmittingCombResolver getSubmittingCombResolver() {
		return submittingCombResolver;
	}

	protected ObjectInvoker getObjectInvoker() {
		return objectInvoker;
	}

	protected BinderProvider getBinderProvider() {
		return binderProvider;
	}

	protected BindablesFactory getBindablesFactory() {
		return bindablesFactory;
	}

	protected HoneycombConverter getHoneycombConverter() {
		return honeycombConverter;
	}

	protected CombRegistry getCombRegistry() {
		return combRegistry;
	}

	protected ControllerProvider getControllerProvider() {
		return controllerProvider;
	}

	protected ConstraintProvider getConstraintProvider() {
		return constraintProvider;
	}

	protected RequestToCombIdTranslator getRequestToCombIdTranslator() {
		return requestToCombIdTranslator;
	}

	public void addConstraint(final String id, final Object constraint) {
		constraintProvider.addConstraint(id, constraint);
	}

	public void addController(final String id, final Object controller) {
		controllerProvider.addController(id, controller);
	}

	public void addBinder(final String id, final Object binder) {
		binderProvider.addBinder(id, binder);
	}

}
